#ifndef _BABELTRACE_TRACE_COLLECTION_H
#define _BABELTRACE_TRACE_COLLECTION_H
/*
 * BabelTrace lib
 *
 * trace collection header
 *
 * Copyright 2012 EfficiOS Inc. and Linux Foundation
 *
 * Author: Mathieu Desnoyers <mathieu.desnoyers@efficios.com>
 * Author: Yannick Brosseau <yannick.brosseau@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifdef __cplusplus
extern "C" {
#endif

struct trace_collection;

void bt_init_trace_collection(struct trace_collection *tc);
void bt_finalize_trace_collection(struct trace_collection *tc);
int bt_trace_collection_add(struct trace_collection *tc,
			 struct bt_trace_descriptor *td);
int bt_trace_collection_remove(struct trace_collection *tc,
			 struct bt_trace_descriptor *td);

#ifdef __cplusplus
}
#endif

#endif /* _BABELTRACE_TRACE_COLLECTION_H */
